module.exports = function() {

  const EventEmitter = require('events');
  const watchman = require('fb-watchman');
  const path = require('path');

  class Watchman extends EventEmitter {

    constructor(opts) {
      super();
      this.opts = opts;
      this.subscription = `sync-${~~(Math.random() * 1000)}`;
    }

    get _client() {
      if(!this.__client) {
        this.__client = new watchman.Client();
      }
      return this.__client;
    }

    _warning(...args) {
      console.log('Watchman:', ...args);
    }

    _capabilities() {
      return new Promise((resolve, reject) => {
        this._client.capabilityCheck({ optional: [], required: [ 'relative_root' ] }, (error, resp) => {
            if(error) {
              this._client.end();
              return reject(error);
            }
            resolve(resp);
          });
      });
    }

    _watch() {
      let source = this.opts.source;
      return new Promise((resolve, reject) => {
        this._client.command([ 'watch-project', source ], (error, resp) => {
          if(error) {
            return reject(error);
          }

          if('warning' in resp) {
            this._warning(resp.warning);
          }

          this._ctx = resp;

          resolve(resp);
        });
      });
    }

    _clock() {
      return new Promise((resolve, reject) => {
      	let { watch } = this._ctx;
      	this._client.command([ 'clock', watch ], (error, resp) => {
  				if(error) {
  				      return reject(error);
  				}
  				this.opts.since = resp.clock;
  				resolve(resp);
  			});
  		});
    }

    _subscribe() {
      let { watch } = this._ctx;
      let { source, since, exclude } = this.opts;
      let subscription = this.subscription;

      let expression = [ 'allof', [ 'type', 'f' ] ];
      for(let path of exclude) {
        expression.push([
          'not', [
            'anyof',
            [ 'match', path, 'basename' ],
            [ 'match', `${path}/**/*`, 'wholename', { includedotfiles: true } ],
            [ 'match', `*/${path}/**/*`, 'wholename', { includedotfiles: true } ],
          ]
        ]);
      }

      let opts = {
        expression,
        fields: [ 'name', 'exists' ],
        since
      };

      return new Promise((resolve, reject) => {
        this._client.command([ 'subscribe', watch, subscription, opts ], (error, resp) => {
          if(error) {
            return reject(error);
          }
          this._client.on('subscription', info => this._onChange(info));
          resolve();
        });
      });
    }

    _onChange(info) {
      if(info.subscription !== this.subscription) {
        return;
      }
      let base = `${path.basename(this.opts.source)}/`;
      this.emit('change', info.files.map(file => {
        let name = file.name;
        if(name.startsWith(base)) {
          file.name = name.substring(base.length);
        }
        return file;
      }));
    }

    start() {
      return this._capabilities()
        .then(() => this._watch())
        .then(() => this._clock())
        .then(() => this._subscribe());
    }
  }

  return Watchman;
}();
